from collections import defaultdict
from typing import Any, Dict, List, Tuple

from sortedcontainers import SortedKeyList

from galahad.server.dataclasses import Annotation, Document


class Annotations:
    def __init__(self, text: str):
        self._text = text
        self._index: Dict[str, SortedKeyList] = defaultdict(lambda: SortedKeyList(key=_sort_func))

    @staticmethod
    def from_dict(text: str, annotations: Dict[str, List[Annotation]]) -> "Annotations":
        result = Annotations(text)

        for type_name, annotations_for_type in annotations.items():
            result._index[type_name].update(annotations_for_type)

        return result

    @staticmethod
    def from_document(document: Document) -> "Annotations":
        result = Annotations(document.text)

        for type_name, annotations in document.annotations.items():
            for annotation in annotations:
                result.create_annotation(type_name, annotation.begin, annotation.end, annotation.features)

        return result

    def to_dict(self) -> Dict[str, List[Dict[str, Any]]]:
        result = defaultdict(list)
        for type_name, annotations in self._index.items():
            for annotation in annotations:
                result[type_name].append(
                    {"begin": annotation.begin, "end": annotation.end, "features": annotation.features}
                )

        return result

    def get_annotations(self) -> Dict[str, List[Annotation]]:
        return {type_name: list(annos) for type_name, annos in self._index.items()}

    def create_annotation(self, type_name, begin: int, end: int, features: Dict[str, Any] = None) -> Annotation:
        if features is None:
            features = {}

        annotation = Annotation(begin=begin, end=end, features=features)
        self._index[type_name].add(annotation)
        return annotation

    def get_covered_text(self, annotation: Annotation) -> str:
        return self._text[annotation.begin : annotation.end]

    def select(self, type_name: str) -> List[Annotation]:
        return list(self._index[type_name])

    def select_covered(self, type_name: str, covering_annotation: Annotation) -> List[Annotation]:
        """Returns a list of covered annotations.

        Return all annotations that are covered
        Only returns annotations that are fully covered, overlapping annotations
        are ignored.

        Args:
            type_name: The type name of the annotations to be returned.
            covering_annotation: The name of the annotation which covers.

        Returns:
            A list of covered annotations
        """
        c_begin = covering_annotation.begin
        c_end = covering_annotation.end

        result = []
        for annotation in self._get_feature_structures_in_range(type_name, c_begin, c_end):
            if annotation.begin >= c_begin and annotation.end <= c_end:
                result.append(annotation)
        return result

    def _get_feature_structures_in_range(self, type_name: str, begin: int, end: int) -> List[Annotation]:
        """Returns a list of all feature structures of type `type_name`.

        Only features are returned that are in [begin, end] or close to it. If you use this function,
        you should always check bound in the calling method.
        """

        annotations = self._index[type_name]

        # We use binary search to find indices for the first and last annotations that are inside
        # the window of [begin, end].
        idx_begin = annotations.bisect_key_left((begin, begin))
        idx_end = annotations.bisect_key_right((end, end))

        return annotations[idx_begin:idx_end]

    @property
    def text(self) -> str:
        return self._text


def _sort_func(a: Annotation) -> Tuple[int, int]:
    return a.begin, a.end



import asyncio
import pathlib
import re
import shutil
from concurrent.futures import ProcessPoolExecutor
from typing import Callable

from fastapi import FastAPI, HTTPException, Path, Response, status
from starlette.background import BackgroundTasks

from galahad.server.classifier import (Classifier, ClassifierStore,
                                       train_classifier)
from galahad.server.dataclasses import *
from galahad.server.util import (get_dataset_folder, get_datasets_folder,
                                 get_document_path)

# This regex forbids two consecutive dots so that ../foo does not work
# to discovery files outside of the document folder
PATH_REGEX = r"^[a-zA-Z0-9_]+(?:\.[a-zA-Z0-9_]+)*$"


def check_naming_is_ok_regex(name: str):
    if not re.match(PATH_REGEX, name):
        raise ValueError(f'The name "{name}" is invalid. ' "Please look at the documentation for correct naming.")


class GalahadServer(FastAPI):
    """Creates a Galahad server instance."""

    def __init__(self, title: str = "Galahad Server", data_dir: pathlib.Path = None) -> None:
        """Creates a Galahad server instance."""
        super().__init__(title=title)

        if data_dir is None:
            data_dir = pathlib.Path.cwd() / "galahad_data"

        data_dir.mkdir(exist_ok=True, parents=True)
        get_datasets_folder(data_dir).mkdir(exist_ok=True, parents=True)
        get_datasets_folder(data_dir).mkdir(exist_ok=True, parents=True)

        self._classifier_store = ClassifierStore(data_dir / "models")

        self.state.data_dir = data_dir
        self.state.lock_dir = data_dir / "locks"
        self.state.classifier_store = self._classifier_store

        _register_routes(self)

    def add_classifier(self, name: str, classifier: Classifier):
        check_naming_is_ok_regex(name)

        document_path = get_document_path(self.state.data_dir, "classifier", name)
        document_path.unlink(missing_ok=True)
        self._classifier_store.add_classifier(name, classifier)


def _register_routes(app: FastAPI):
    data_dir: pathlib.Path = app.state.data_dir
    lock_directory = app.state.lock_dir
    classifier_store: ClassifierStore = app.state.classifier_store

    # Scheduling
    # https://stackoverflow.com/questions/63169865/how-to-do-multiprocessing-in-fastapi

    @app.on_event("startup")
    async def startup_event():
        app.state.executor = ProcessPoolExecutor()

    @app.on_event("shutdown")
    async def on_shutdown():
        app.state.executor.shutdown()

    async def run_in_different_process(fn: Callable, *args):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(app.state.executor, fn, *args)

    # Meta

    @app.get("/ping")
    def ping():
        return {"ping": "pong"}

    # Dataset

    @app.get(
        "/dataset",
        response_model=DatasetList,
        responses={status.HTTP_200_OK: {"description": "Returns list of all available datasets."}},
        status_code=status.HTTP_200_OK,
    )
    def list_datasets():
        """Lists dataset names managed by this server."""
        dataset_names = []

        for p in sorted(get_datasets_folder(data_dir).iterdir()):
            dataset_names.append(p.name)

        return DatasetList(names=dataset_names)

    @app.put(
        "/dataset/{dataset_id}",
        responses={
            status.HTTP_204_NO_CONTENT: {"description": "Dataset created."},
            status.HTTP_409_CONFLICT: {"description": "Dataset already exists."},
        },
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def create_dataset(
        dataset_id: str = Path(..., title="Identifier of the dataset that should be created", regex=PATH_REGEX),
    ):
        """Creates a dataset with the given `dataset_id`. Does nothing and returns `409` if it already existed."""
        dataset_folder = get_dataset_folder(data_dir, dataset_id)

        if dataset_folder.exists():
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT, detail=f"Dataset with id [{dataset_id}] already exists."
            )

        dataset_folder.mkdir(parents=True)
        return Response(content="", status_code=status.HTTP_204_NO_CONTENT)

    @app.delete(
        "/dataset/{dataset_id}",
        responses={
            status.HTTP_204_NO_CONTENT: {"description": "Dataset deleted."},
            status.HTTP_404_NOT_FOUND: {"description": "Dataset not found."},
        },
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def delete_dataset(
        dataset_id: str = Path(..., title="Identifier of the dataset that should be deleted", regex=PATH_REGEX),
    ):
        """Deletes the dataset with the given `dataset_id` and its documents."""
        dataset_folder = get_dataset_folder(data_dir, dataset_id)

        if not dataset_folder.is_dir():
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Dataset with id [{dataset_id}] not found."
            )

        shutil.rmtree(dataset_folder)

        return Response(content="", status_code=status.HTTP_204_NO_CONTENT)

    @app.get(
        "/dataset/{dataset_id}",
        response_model=DocumentList,
        responses={
            status.HTTP_200_OK: {"description": "Returns list of documents in dataset."},
            status.HTTP_404_NOT_FOUND: {"description": "Dataset not found."},
        },
        status_code=status.HTTP_200_OK,
    )
    def list_documents_in_dataset(
        dataset_id: str = Path(
            ..., title="Identifier of the dataset whose documents should be listed", regex=PATH_REGEX
        ),
    ):
        """Lists documents in the dataset with the given `dataset_id`."""
        dataset_folder = get_dataset_folder(data_dir, dataset_id)

        if not dataset_folder.is_dir():
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Dataset with id [{dataset_id}] not found."
            )

        names = []
        versions = []

        for p in sorted(dataset_folder.iterdir()):
            document: Document = Document.parse_file(p)
            names.append(p.name)
            versions.append(document.version)

        return DocumentList(names=names, versions=versions)

    @app.put(
        "/dataset/{dataset_id}/{document_id}",
        responses={
            status.HTTP_204_NO_CONTENT: {"description": "Document added."},
            status.HTTP_404_NOT_FOUND: {"description": "Dataset not found."},
        },
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def add_document_to_dataset(
        request: Document,
        dataset_id: str = Path(..., title="Identifier of the dataset to add to", regex=PATH_REGEX),
        document_id: str = Path(..., title="Identifier of the document to add", regex=PATH_REGEX),
    ):
        """Adds a document to an already existing dataset. Overwrites a document if it already existed."""
        dataset_folder = get_dataset_folder(data_dir, dataset_id)

        if not dataset_folder.is_dir():
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Dataset with id [{dataset_id}] not found."
            )

        document_path = get_document_path(data_dir, dataset_id, document_id)
        with document_path.open("w", encoding="utf-8") as f:
            f.write(request.json(skip_defaults=True))

        return Response(content="", status_code=status.HTTP_204_NO_CONTENT)

    @app.delete(
        "/dataset/{dataset_id}/{document_id}",
        responses={
            status.HTTP_204_NO_CONTENT: {"description": "Document deleted."},
            status.HTTP_404_NOT_FOUND: {"description": "Dataset not found."},
        },
        status_code=status.HTTP_204_NO_CONTENT,
    )
    def delete_document_from_dataset(
        dataset_id: str = Path(..., title="Identifier of the dataset to delete from", regex=PATH_REGEX),
        document_id: str = Path(..., title="Identifier of the document to delete", regex=PATH_REGEX),
    ):
        """Deletes a document from a dataset. Does nothing if the document did not exist."""
        dataset_folder = get_dataset_folder(data_dir, dataset_id)

        if not dataset_folder.is_dir():
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Dataset with id [{dataset_id}] not found."
            )

        document_path = get_document_path(data_dir, dataset_id, document_id)
        document_path.unlink(missing_ok=True)

        return Response(content="", status_code=status.HTTP_204_NO_CONTENT)

    # Model

    @app.get(
        "/classifier",
        response_model=List[ClassifierInfo],
        responses={
            status.HTTP_200_OK: {"description": "Returns list of documents in dataset."},
        },
        status_code=status.HTTP_200_OK,
    )
    def get_all_classifier_infos():
        """Gets the classifier info for all classifiers managed by this server."""
        return classifier_store.get_classifier_infos()

    @app.get(
        "/classifier/{classifier_id}",
        response_model=ClassifierInfo,
        responses={
            status.HTTP_200_OK: {"description": "Returns the classifier info of the requested classifier."},
            status.HTTP_404_NOT_FOUND: {"description": "Classifier not found."},
        },
        status_code=status.HTTP_200_OK,
    )
    def get_classifier_info(
        classifier_id: str = Path(..., title="Identifier of the classifier whose info to query", regex=PATH_REGEX)
    ):
        """Gets the classifier info for the requested classifier id if it exists."""
        classifier_info = classifier_store.get_classifier_info(classifier_id)

        if classifier_info is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Classifier with id [{classifier_id}] not found."
            )

        return classifier_info

    # Train

    @app.post(
        "/classifier/{classifier_id}/{model_id}/train/{dataset_id}",
        responses={
            status.HTTP_202_ACCEPTED: {"description": "Training started."},
            status.HTTP_404_NOT_FOUND: {"description": "Classifier or dataset not found."},
            status.HTTP_429_TOO_MANY_REQUESTS: {"description": "Already training."},
        },
    )
    def train_on_dataset(
        background_tasks: BackgroundTasks,
        classifier_id: str = Path(..., title="Name of the classifier that should be trained.", regex=PATH_REGEX),
        model_id: str = Path(..., title="Name of the model that should be trained.", regex=PATH_REGEX),
        dataset_id: str = Path(
            ..., title="Identifier of the dataset that should be used for training", regex=PATH_REGEX
        ),
    ):
        classifier = classifier_store.get_classifier(classifier_id)
        if classifier is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Classifier with id [{classifier_id}] not found."
            )

        dataset_folder = get_dataset_folder(data_dir, dataset_id)
        if not dataset_folder.is_dir():
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Dataset with id [{dataset_id}] not found."
            )

        background_tasks.add_task(
            run_in_different_process, train_classifier, classifier, dataset_folder, model_id, lock_directory
        )

        return Response(content="", status_code=status.HTTP_202_ACCEPTED)

    # Prediction

    @app.post(
        "/classifier/{classifier_id}/{model_id}/predict",
        response_model=Document,
        responses={
            status.HTTP_200_OK: {"description": "Prediction successful."},
            status.HTTP_404_NOT_FOUND: {"description": "Classifier or model not found."},
        },
    )
    def predict_for_document(
        request: Document,
        classifier_id: str = Path(
            ..., title="Name of the classifier that should be used for prediction", regex=PATH_REGEX
        ),
        model_id: str = Path(..., title="Identifier of the model that should be used for prediction", regex=PATH_REGEX),
    ):
        classifier = classifier_store.get_classifier(classifier_id)
        if classifier is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail=f"Classifier with id [{classifier_id}] not found."
            )

        result = classifier.predict(model_id, request)
        if result is None:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"Model with id [{model_id}] not found.")

        return result

    @app.post("/classifier/{classifier_id}/{model_id}/predict/{dataset_id}/{document_id}")
    def predict_on_dataset(
        classifier_id: str = Path(..., title="Identifier of the model that should be used for prediction"),
        dataset_id: str = Path(..., title="Identifier of the dataset that should be used for prediction"),
        document_id: str = Path(
            ..., title="Identifier of the document in the given dataset that should be used for prediction"
        ),
    ):
        return {"classifier_id": classifier_id, "dataset_id": dataset_id, "document_id": document_id}



from typing import Any, Dict, List

from pydantic import BaseModel, Field


class Annotation(BaseModel):
    begin: int
    end: int
    features: Dict[str, Any] = Field(default_factory=dict)


Layer = List[Annotation]
Layers = Dict[str, Layer]


# Datasets


class DatasetList(BaseModel):
    names: List[str]

    class Config:
        schema_extra = {"example": {"names": ["dataset1", "dataset2", "dataset3"]}}


class Document(BaseModel):
    text: str  # Document text
    annotations: Dict[
        str, Layer
    ]  # The annotations in the document, one dict per type, start and end offsets index into `text`
    version: int = Field(default=0)  # Version of the document, needs to be monotonically increasing

    class Config:
        schema_extra = {
            "example": {
                "text": "Joe waited for the train . The train was late .",
                "version": 23,
                "annotations": {
                    "t.token": [
                        {"begin": 0, "end": 3},
                        {"begin": 4, "end": 10},
                        {"begin": 11, "end": 14},
                        {"begin": 15, "end": 18},
                        {"begin": 19, "end": 24},
                        {"begin": 25, "end": 26},
                        {"begin": 27, "end": 30},
                        {"begin": 31, "end": 36},
                        {"begin": 37, "end": 40},
                        {"begin": 41, "end": 45},
                        {"begin": 46, "end": 47},
                    ],
                    "t.sentence": [
                        {"begin": 0, "end": 26},
                        {"begin": 27, "end": 47},
                    ],
                    "t.named_entity": [
                        {"begin": 0, "end": 3, "features": {"f.value": "PER"}},
                    ],
                },
            }
        }


class DocumentList(BaseModel):
    names: List[str]
    versions: List[int]

    class Config:
        schema_extra = {
            "example": {"names": ["document1.xmi", "document2.txt", "document3.pdf"], "versions": [7, 6, 7]}
        }


# Classifier


class ClassifierInfo(BaseModel):
    name: str

    class Config:
        schema_extra = {"example": {"name": "ExampleClassifier"}}


# Training


class TrainingRequest(BaseModel):
    metadata: Dict[str, Any]


# Predicting


class PredictionRequest(BaseModel):
    metadata: Dict[str, Any]
    text: str
    data: Dict[str, Layer]

    class Config:
        schema_extra = {
            "example": {
                "metadata": {
                    "token_type": "t.token",
                    "target_type": "t.ner",
                    "target_feature": "label",
                },
                "text": "Joe waited for the train . The train was late .",
                "data": {
                    "t.token": [
                        {"begin": 0, "end": 3},
                        {"begin": 4, "end": 10},
                        {"begin": 11, "end": 14},
                        {"begin": 15, "end": 18},
                        {"begin": 19, "end": 24},
                        {"begin": 25, "end": 26},
                        {"begin": 27, "end": 30},
                        {"begin": 31, "end": 36},
                        {"begin": 37, "end": 40},
                        {"begin": 41, "end": 45},
                        {"begin": 46, "end": 47},
                    ],
                    "t.sentence": [
                        {"begin": 0, "end": 26},
                        {"begin": 27, "end": 47},
                    ],
                },
            }
        }


class PredictionResponse(BaseModel):
    data: Dict[str, Layer]

    class Config:
        schema_extra = {
            "example": {
                "data": {
                    "t.ner": [
                        {"begin": 0, "end": 3},
                    ]
                }
            }
        }



import logging
import os
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

import joblib
from filelock import FileLock

from galahad.server.dataclasses import ClassifierInfo, Document

logger = logging.getLogger(__file__)


class AnnotationTypes(Enum):
    TOKEN = "t.token"
    SENTENCE = "t.sentence"
    ANNOTATION = "t.annotation"


class AnnotationFeatures(Enum):
    VALUE = "f.value"


class Remapper:
    def __init__(self, remaps: Dict[str, str] = None):
        if remaps is None:
            remaps = {}

        self._remaps = remaps

    def remap(self, name: str) -> str:
        return self._remaps.get(name, name)


class Classifier:
    def __init__(self):
        self._model_directory: Optional[Path] = None

    def train(self, model_id: str, documents: List[Document]):
        pass

    def predict(self, model_id: str, document: Document) -> Optional[Document]:
        raise NotImplementedError()

    def consumes(self) -> List[str]:
        return []

    def produces(self) -> List[str]:
        return []

    def _save_model(self, model_id: str, model: Any):
        assert self._model_directory is not None, "You need to set `self._model_directory` first before saving!"

        model_path = self._get_model_path(model_id)
        model_path.parent.mkdir(parents=True, exist_ok=True)

        tmp_model_path = model_path.with_suffix(".joblib.tmp")
        joblib.dump(model, tmp_model_path)

        os.replace(tmp_model_path, model_path)

    def _load_model(self, model_id: str) -> Optional[Any]:
        model_path = self._get_model_path(model_id)
        if model_path.is_file():
            logger.debug("Model found for [%s]", model_path)
            return joblib.load(model_path)
        else:
            logger.debug("No model found for [%s]", model_path)
            return None

    def _get_model_path(self, model_id: str) -> Path:
        return self._model_directory / self.name / f"model_{model_id}.joblib"

    @property
    def name(self) -> str:
        return type(self).__name__


class ClassifierStore:
    def __init__(self, model_directory: Path):
        self._model_directory = model_directory
        self._classifiers: Dict[str, Classifier] = {}

    def add_classifier(self, name: str, classifier: Classifier):
        if name in self._classifiers:
            raise ValueError(f"Model [{name}] already in classifier store!")

        classifier._model_directory = self._model_directory
        self._classifiers[name] = classifier

    def get_classifier(self, name: str) -> Optional[Classifier]:
        return self._classifiers.get(name)

    def get_classifier_info(self, name: str) -> Optional[ClassifierInfo]:
        """Builds classifier info for the classifier given by `name` and returns it.

        Args:
            name: The name of the classifier whose info to get.

        Returns:
            The classifier info of the classifier named `name` if it was found, else `None`.
        """
        classifier = self._classifiers.get(name)
        if not classifier:
            return None

        return ClassifierInfo(name=name)

    def get_classifier_infos(self) -> List[ClassifierInfo]:
        """Builds classifier infos for all classifiers in this store and returns it.

        Returns:
            List of classifier infos for all stored classifiers.
        """
        return [self.get_classifier_info(name) for name in sorted(self._classifiers.keys())]


def train_classifier(classifier: Classifier, dataset_folder: Path, model_id: str, lock_directory: Path):
    lock = get_lock(lock_directory, model_id)

    try:
        lock.acquire()

        documents = [Document.parse_file(p) for p in sorted(dataset_folder.iterdir())]
        classifier.train(model_id, documents)
    except TimeoutError:
        logger.info("Already training [%s] with model id [%s], skipping!", classifier.name, model_id)
    finally:
        lock.release()


def get_lock(lock_directory: Path, lock_id: str) -> FileLock:
    lock_directory.mkdir(parents=True, exist_ok=True)
    lock_path = lock_directory / f"{lock_id}.lock"
    return FileLock(str(lock_path), timeout=1)



import logging
from typing import Dict, List

import requests
from requests_toolbelt import sessions

from galahad.server import server
from galahad.server.dataclasses import ClassifierInfo, Document

logger = logging.getLogger("galahad.client")


class HTTPError(Exception):
    def __init__(self, response: requests.Response):
        error_msg = ""

        if isinstance(response.reason, bytes):
            # We attempt to decode utf-8 first because some servers
            # choose to localize their reason strings. If the string
            # isn't utf-8, we fall back to iso-8859-1 for all other
            # encodings. (See PR #3538)
            try:
                response_reason = response.reason
                reason = response_reason.decode("utf-8")
            except UnicodeDecodeError:
                reason = response.reason.decode("iso-8859-1")
        else:
            reason = response.reason

        status_code = response.status_code
        if 400 <= status_code < 500:
            error_msg = f"{status_code} Client Error: {reason} for url: {response.url}"

        elif 500 <= status_code < 600:
            error_msg = f"{status_code} Server Error: {reason} for url: {response.url}"

        body = response.content
        if len(body):
            error_msg += "\n"
            error_msg += body.decode("utf-8")

        super().__init__(error_msg)


def create_error_message(**kwargs) -> str:
    variables = kwargs
    if len(variables.keys()) == 0:
        return ""
    else:
        error_message = ["The problem appeared with the variables "]
        for variable_name in variables.keys():
            # variable_list.append(variable_name)
            # variable_list.append(variables[variable_name])
            error_message.append(f'"{variable_name}"')
            error_message.append(":")
            error_message.append(f'"{variables[variable_name]}"')
            error_message.append("and")
        del error_message[-1]

    return " ".join(error_message)


def check_response(response: requests.Response):
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise HTTPError(response) from e


# TODO: would it be better for the performance if we checked the naming before sending the request to the server?
def check_naming_is_ok(given_status: int, **kwargs):
    if given_status == 422:
        raise ValueError(
            f"Naming for one of the variables is invalid. "
            f"Please look at the documentation for correct naming. {create_error_message(**kwargs)}"
        )


class GalahadClient:
    def __init__(self, endpoint_url: str):
        self.endpoint_url = endpoint_url.rstrip("/")
        self._session = self._build_session()

    def start_session(self) -> requests.Session:
        self._session = self._build_session()
        return self._session

    def _build_session(self) -> requests.Session:
        session = sessions.BaseUrlSession(self.endpoint_url)
        return session

    def is_connected(self) -> bool:
        response = self._session.get("/ping")
        if response.status_code != 200:
            logger.info("StatusCodeError")
            return False
        if response.json() != {"ping": "pong"}:
            logger.info("ResponseError")
            return False
        return True

    # output is sorted by dataset name
    def list_datasets(self) -> List[str]:
        response = self._session.get("/dataset")
        check_response(response)
        return response.json()["names"]

    def contains_dataset(self, dataset_id: str) -> bool:
        server.check_naming_is_ok_regex(dataset_id)
        return dataset_id in self.list_datasets()

    def create_dataset(self, dataset_id: str):
        response = self._session.put(f"/dataset/{dataset_id}", {})
        check_naming_is_ok(response.status_code, dataset_id=dataset_id)
        if response.status_code == 409:
            logger.info(f'Dataset with id "{dataset_id}" already exists')
            return None

        check_response(response)

    def delete_dataset(self, dataset_id: str):
        response = self._session.delete(f"/dataset/{dataset_id}")
        check_naming_is_ok(response.status_code, dataset_id=dataset_id)
        if response.status_code == 404:
            logger.info(f'Dataset with id "{dataset_id}" does not exist')
            return None

        check_response(response)

    def delete_datasets(self, dataset_ids: List[str]):
        for dataset_id in dataset_ids:
            self.delete_dataset(dataset_id)

    def delete_all_datasets(self):
        self.delete_datasets(self.list_datasets())

    # The new document of the same name will override an existing one!
    def create_document_in_dataset(
        self, dataset_id: str, document_id: str, document: Document, auto_create_dataset=False
    ):
        response = self._session.put(f"/dataset/{dataset_id}/{document_id}", json=document.dict())
        check_naming_is_ok(response.status_code, dataset_id=dataset_id, document_id=document_id)

        if response.status_code == 404:
            if auto_create_dataset:
                self.create_dataset(dataset_id)
                response = self._session.put(f"/dataset/{dataset_id}/{document_id}", json=document.dict())
            else:
                raise ValueError(
                    f'The dataset for the given id: "{dataset_id}" does not exist. To create it, '
                    'set the optional parameter "auto_create_dataset" to True'
                )

        check_response(response)

    # result is sorted by doc id
    def list_documents_in_dataset(self, dataset_id) -> Dict[str, int]:
        response = self._session.get(f"/dataset/{dataset_id}")
        check_naming_is_ok(response.status_code, dataset_id=dataset_id)
        check_response(response)

        return dict(zip(response.json()["names"], response.json()["versions"]))

    def dataset_contains_document(self, dataset_id: str, document_id: str) -> bool:
        server.check_naming_is_ok_regex(document_id)
        return document_id in list(self.list_documents_in_dataset(dataset_id).keys())

    def delete_document_in_dataset(self, dataset_id: str, document_id: str):
        if self.dataset_contains_document(dataset_id, document_id):
            response = self._session.delete(f"/dataset/{dataset_id}/{document_id}")
            check_naming_is_ok(response.status_code, dataset_id=dataset_id, document_id=document_id)
            check_response(response)
        else:
            logger.info(f'Document with id "{document_id}" does not exist in dataset with id "{dataset_id}"')

    def delete_all_documents_in_dataset(self, dataset_id: str):
        for document_id in list(self.list_documents_in_dataset(dataset_id).keys()):
            self.delete_document_in_dataset(dataset_id, document_id)

    def delete_all_documents(self):
        for dataset_id in self.list_datasets():
            self.delete_all_documents_in_dataset(dataset_id)

    def list_all_classifiers(self) -> List[ClassifierInfo]:
        response = self._session.get("/classifier")
        check_response(response)

        info_list = []
        for classifier in response.json():
            info_list.append(ClassifierInfo.parse_obj(classifier))

        return info_list

    def get_classifier_info(self, classifier_id: str) -> ClassifierInfo:
        response = self._session.get(f"/classifier/{classifier_id}")
        check_naming_is_ok(response.status_code, classifier_id=classifier_id)
        check_response(response)

        return ClassifierInfo.parse_obj(response.json())

    # True: training has started. False: training has started already and function call had no effect
    def train_on_dataset(self, classifier_id: str, model_id: str, dataset_id: str) -> bool:
        response = self._session.post(f"/classifier/{classifier_id}/{model_id}/train/{dataset_id}")
        check_naming_is_ok(response.status_code, classifier_id=classifier_id, model_id=model_id, dataset_id=dataset_id)
        if response.status_code == 429:
            # logger.info("Training has already started! {create_error_message(variables)}")
            return False

        check_response(response)

        return True

    def predict_on_document(self, classifier_id: str, model_id: str, document: Document) -> Document:
        response = self._session.post(
            f"{self.endpoint_url}/classifier/{classifier_id}/{model_id}/predict", json=document.dict()
        )

        check_naming_is_ok(response.status_code, classifier_id=classifier_id, model_id=model_id)
        check_response(response)
        return response.json()



import logging
from typing import List, Optional

try:
    from sklearn.feature_extraction.text import (CountVectorizer,
                                                 TfidfTransformer)
    from sklearn.naive_bayes import MultinomialNB
    from sklearn.pipeline import Pipeline
except ImportError as error:
    print("Could not import 'sklearn', please install it manually via 'pip install scikit-learn'")

from galahad.formats import build_sentence_classification_document
from galahad.server.annotations import Annotations
from galahad.server.classifier import (AnnotationFeatures, AnnotationTypes,
                                       Classifier)
from galahad.server.dataclasses import Document

logger = logging.getLogger(__name__)


class SklearnSentenceClassifier(Classifier):
    def __init__(self):
        super().__init__()

        self._sentence_type = AnnotationTypes.SENTENCE.value
        self._sentence_annotation_type = AnnotationTypes.ANNOTATION.value
        self._target_feature = AnnotationFeatures.VALUE.value

    def train(self, model_id: str, documents: List[Document]):
        texts = []
        labels = []

        for document in documents:
            annotations = Annotations.from_dict(document.text, document.annotations)

            for sentence in annotations.select(self._sentence_type):
                for sentence_label in annotations.select_covered(self._sentence_annotation_type, sentence):
                    text = annotations.get_covered_text(sentence_label)
                    label = sentence_label.features.get(self._target_feature)

                    if label is None:
                        continue

                    texts.append(text)
                    labels.append(label)

        assert len(texts) == len(labels), "Unequal number of sentences and labels"
        if not len(texts):
            logger.debug(f"Empty training set, skipping!")

        model = Pipeline([("vect", CountVectorizer()), ("tfidf", TfidfTransformer()), ("clf", MultinomialNB())])
        model.fit(texts, labels)

        logger.debug(f"Training finished for model with id [%s]", model_id)

        self._save_model(model_id, model)

    def predict(self, model_id: str, document: Document) -> Optional[Document]:
        model: Optional[Pipeline] = self._load_model(model_id)

        if model is None:
            logger.debug("No trained model ready yet!")
            return

        annotations = Annotations.from_dict(document.text, document.annotations)
        texts = [annotations.get_covered_text(sentence) for sentence in annotations.select(self._sentence_type)]
        predicted_labels = model.predict(texts)

        return build_sentence_classification_document(texts, predicted_labels)

    def consumes(self) -> List[str]:
        return [self._sentence_type, self._sentence_annotation_type]

    def produces(self) -> List[str]:
        return [self._target_feature]



from typing import Optional

try:
    import spacy as spacy
    from spacy.tokens import Doc
except ImportError as error:
    print("Could not import 'spacy', please install it manually via 'pip install spacy'")

from galahad.formats import Span, build_span_classification_response
from galahad.server.annotations import Annotations
from galahad.server.classifier import (AnnotationFeatures, AnnotationTypes,
                                       Classifier)
from galahad.server.dataclasses import Document


class SpacyNerTagger(Classifier):
    def __init__(self, model_name: str):
        super().__init__()

        self._token_type = AnnotationTypes.TOKEN.value
        self._target_feature = AnnotationFeatures.VALUE.value

        self._model = spacy.load(model_name, disable=["parser"])

    def predict(self, model_id: str, document: Document) -> Optional[Document]:
        # Extract the tokens from the document and create a spacy doc from it
        annotations = Annotations.from_dict(document.text, document.annotations)
        words = [annotations.get_covered_text(token) for token in annotations.select(self._token_type)]

        doc = Doc(self._model.vocab, words=words)

        # Find the named entities
        self._model.get_pipe("ner")(doc)

        # For every entity returned by spacy, create an annotation in the resulting doc
        spans = []
        for named_entity in doc.ents:
            spans.append(Span(named_entity.start, named_entity.end, named_entity.label_))

        return build_span_classification_response(document, spans)



from typing import Optional

try:
    import spacy as spacy
    from spacy.tokens import Doc
except ImportError as error:
    print("Could not import 'spacy', please install it manually via 'pip install spacy'")

from galahad.formats import build_token_labeling_response
from galahad.server.annotations import Annotations
from galahad.server.classifier import (AnnotationFeatures, AnnotationTypes,
                                       Classifier)
from galahad.server.dataclasses import Document


class SpacyPosTagger(Classifier):
    def __init__(self, model_name: str):
        super().__init__()

        self._token_type = AnnotationTypes.TOKEN.value
        self._target_feature = AnnotationFeatures.VALUE.value

        self._model = spacy.load(model_name, disable=["parser"])

    def predict(self, model_id: str, document: Document) -> Optional[Document]:
        # Extract the tokens from the document and create a spacy doc from it
        annotations = Annotations.from_dict(document.text, document.annotations)
        words = [annotations.get_covered_text(token) for token in annotations.select(self._token_type)]

        spacy_doc = Doc(self._model.vocab, words=words)

        self._model.get_pipe("tok2vec")(spacy_doc)
        self._model.get_pipe("tagger")(spacy_doc)

        list_of_pos_tags = []
        for i in range(len(spacy_doc)):
            list_of_pos_tags.append(spacy_doc[i].tag_)

        return build_token_labeling_response(document, list_of_pos_tags)



from pathlib import Path


def get_datasets_folder(data_dir: Path) -> Path:
    result = data_dir / "datasets"
    if not path_is_parent(data_dir, result):
        raise ValueError("Naming is invalid. Please look at the documentation for correct naming.")
    return result


def get_dataset_folder(data_dir: Path, dataset_id: str) -> Path:
    result = get_datasets_folder(data_dir) / dataset_id
    if not path_is_parent(data_dir, result):
        raise ValueError(
            'Naming for the dataset "' + dataset_id + '" is invalid. '
            "Please look at the documentation for correct naming."
        )
    return result


def get_document_path(data_dir: Path, dataset_id: str, document_name: str) -> Path:
    result = get_dataset_folder(data_dir, dataset_id) / document_name
    if not path_is_parent(data_dir, result):
        raise ValueError(
            'Naming for the dataset "' + dataset_id + '" or for the document "' + document_name + '" is invalid. '
            "Please look at the documentation for correct naming."
        )
    return result


def path_is_parent(parent: Path, child: Path) -> bool:
    """
    Checks whether `parent` is really a parent of `child`.
    This is mainly used to prevent clients sending path names like ../../bar that can
    potentially escape the data folder and cause harm.
    """
    return parent.resolve() in child.resolve().parents



# !/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import find_packages, setup

NAME = "galahad"
DESCRIPTION = "Machine learning model server that can predict AND train"
HOMEPAGE = "https://github.com/inception-project/inception-external-recommender-v2"
EMAIL = "git@mrklie.com"
AUTHOR = "Jan-Christoph Klie"
REQUIRES_PYTHON = ">=3.7.0"

install_requires = [
    "fastapi==0.75.*",
    "uvicorn[standard]==0.17.*",
    "sortedcontainers>==2.4.*",
    "joblib==1.1.*",
    "filelock==3.6.*",
    "requests==2.27.*",
    "requests-toolbelt==0.9.*",
]

demo_dependencies = ["gradio==2.9.*", "nltk==3.7"]

test_dependencies = ["pytest", "datasets"]

dev_dependencies = ["black", "isort"]

doc_dependencies = [
    "sphinx",
    "sphinx-autodoc-typehints",
    "sphinx-rtd-theme",
    "sphinx-autoapi",
    "m2r2",
    "sphinx_copybutton",
    "sphinx_rst_builder",
]

spacy_dependencies = [
    "spacy==3.2.*",
]

sklearn_dependencies = ["scikit-learn>=0.24.*"]

contrib_dependencies = []
contrib_dependencies.extend(spacy_dependencies)
contrib_dependencies.extend(sklearn_dependencies)


extras = {
    "test": test_dependencies,
    "dev": dev_dependencies,
    "doc": doc_dependencies,
    "contrib": contrib_dependencies,
    "spacy": spacy_dependencies,
    "sklearn": sklearn_dependencies,
    "demo": demo_dependencies,
}

all_dependencies = []
for e in extras.values():
    all_dependencies.extend(e)

extras["all"] = all_dependencies

here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r") as fh:
    long_description = fh.read()

# Load the package"s __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, "galahad", "__version__.py")) as f:
    exec(f.read(), about)


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown; charset=UTF-8",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=HOMEPAGE,
    packages=find_packages(exclude="tests"),
    keywords="inception",
    project_urls={
        "Bug Tracker": "https://github.com/inception-project/inception-external-recommender-v2/issues",
        "Documentation": "https://github.com/inception-project/inception-external-recommender-v2",
        "Source Code": "https://github.com/inception-project/inception-external-recommender-v2",
    },
    install_requires=install_requires,
    test_suite="tests",
    tests_require=test_dependencies,
    extras_require=extras,
    include_package_data=True,
    license="Apache License 2.0",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Text Processing :: Linguistic",
    ],
)



import copy
from dataclasses import dataclass
from typing import List

from galahad.server.annotations import Annotations
from galahad.server.classifier import AnnotationFeatures, AnnotationTypes
from galahad.server.dataclasses import Annotation, Document


@dataclass
class Span:
    begin: int
    end: int
    value: str


def build_sentence_classification_document(sentences: List[str], labels: List[str], version: int = 0) -> Document:
    assert len(sentences) == len(labels), "Sentences and labels need to have the same length!"

    text = " ".join(sentences)
    annotations = Annotations(text)

    sentence_type = AnnotationTypes.SENTENCE.value
    sentence_annotation_type = AnnotationTypes.ANNOTATION.value
    value_feature = AnnotationFeatures.VALUE.value

    begin = 0
    for sentence, label in zip(sentences, labels):
        end = begin + len(sentence)

        annotation = annotations.create_annotation(sentence_type, begin, end)
        assert annotations.get_covered_text(annotation) == sentence

        annotation = annotations.create_annotation(sentence_annotation_type, begin, end, {value_feature: label})
        assert annotations.get_covered_text(annotation) == sentence

        begin = end + 1

    document = Document(text=text, annotations=annotations.to_dict(), version=version)
    return document


def build_span_classification_request(
    sentences: List[List[str]], spans: List[List[Span]] = None, version: int = 0
) -> Document:
    text = " ".join(t for sentence in sentences for t in sentence)
    annotations = Annotations(text)

    token_idx_to_begins = {}
    token_idx_to_ends = {}

    token_type = AnnotationTypes.TOKEN.value
    span_annotation_type = AnnotationTypes.ANNOTATION.value
    value_feature = AnnotationFeatures.VALUE.value

    begin = 0
    end = 0
    for sentence_idx, sentence in enumerate(sentences):
        sentence_start = begin
        for token_idx, token_text in enumerate(sentence):
            end = begin + len(token_text)
            token_idx_to_begins[(sentence_idx, token_idx)] = begin
            token_idx_to_ends[(sentence_idx, token_idx)] = end

            annotation = annotations.create_annotation(token_type, begin, end)
            assert annotations.get_covered_text(annotation) == token_text

            begin = end + 1
        annotations.create_annotation(AnnotationTypes.SENTENCE.value, sentence_start, end)

    for sentence_idx, sentence in enumerate(spans or []):
        for span in sentence:
            begin = token_idx_to_begins[(sentence_idx, span.begin)]
            end = token_idx_to_ends[(sentence_idx, span.end - 1)]

            annotations.create_annotation(span_annotation_type, begin, end, {value_feature: span.value})

    document = Document(text=text, annotations=annotations.to_dict(), version=version)
    return document


def build_span_classification_response(original_doc: Document, spans: List[Span] = None, version: int = 0) -> Document:
    annotated_doc = copy.deepcopy(original_doc)
    annotated_doc.version = version

    assert AnnotationTypes.TOKEN.value in original_doc.annotations
    assert AnnotationTypes.SENTENCE.value in original_doc.annotations

    annotations = Annotations.from_dict(annotated_doc.text, annotated_doc.annotations)

    sentences = annotations.select(AnnotationTypes.SENTENCE.value)

    dummy_text_annotation = {"begin": sentences[0].begin, "end": sentences[-1].end}
    all_tokens = annotations.select_covered(AnnotationTypes.TOKEN.value, Annotation(**dummy_text_annotation))
    for span in spans:
        first_token = all_tokens[span.begin]
        last_token = all_tokens[span.end - 1]

        annotations.create_annotation(
            AnnotationTypes.ANNOTATION.value,
            first_token.begin,
            last_token.end,
            {AnnotationFeatures.VALUE.value: span.value},
        )

    annotated_doc.annotations = annotations.get_annotations()
    return annotated_doc


def build_doc_from_tokens_and_text(text: str, sentences: List[List[str]]) -> Document:
    sentence_list = []
    token_list = []
    position = 0
    subtext = text
    sentence_start = 0
    for sentence in sentences:

        for token in sentence:
            start = subtext.find(token)
            position = position + start
            token_list.append(Annotation(**{"begin": position, "end": position + len(token), "features": {}}))
            subtext = subtext[start + len(token) :]
            position = position + len(token)

        sentence_list.append(Annotation(**{"begin": sentence_start, "end": position, "features": {}}))
        sentence_start = position + 1

    doc = Document(text=text, annotations={"t.token": token_list, "t.sentence": sentence_list}, version=0)
    return doc


def build_token_labeling_response(original_doc: Document, labels: List[str] = None, version: int = 0) -> Document:
    annotated_doc = copy.deepcopy(original_doc)
    annotated_doc.version = version

    assert AnnotationTypes.TOKEN.value in original_doc.annotations
    assert AnnotationTypes.SENTENCE.value in original_doc.annotations
    assert len(original_doc.annotations["t.token"]) == len(labels)

    annotations = Annotations.from_dict(annotated_doc.text, annotated_doc.annotations)

    for token, label in zip(original_doc.annotations["t.token"], labels):
        annotations.create_annotation(
            AnnotationTypes.ANNOTATION.value,
            token.begin,
            token.end,
            {AnnotationFeatures.VALUE.value: label},
        )

    annotated_doc.annotations = annotations.get_annotations()
    return annotated_doc


def build_span_classification_response_per_sentence(
    original_doc: Document, spans: List[List[Span]] = None, version: int = 0
) -> Document:
    annotated_doc = copy.deepcopy(original_doc)
    annotated_doc.version = version

    assert AnnotationTypes.TOKEN.value in original_doc.annotations
    assert AnnotationTypes.SENTENCE.value in original_doc.annotations

    annotations = Annotations.from_dict(annotated_doc.text, annotated_doc.annotations)

    sentences = annotations.select(AnnotationTypes.SENTENCE.value)
    assert len(sentences) == len(spans)

    for sentence, cur_spans in zip(sentences, spans):
        tokens = annotations.select_covered(AnnotationTypes.TOKEN.value, sentence)

        for span in cur_spans:
            first_token = tokens[span.begin]
            last_token = tokens[span.end - 1]

            annotations.create_annotation(
                AnnotationTypes.ANNOTATION.value,
                first_token.begin,
                last_token.end,
                {AnnotationFeatures.VALUE.value: span.value},
            )

    annotated_doc.annotations = annotations.get_annotations()
    return annotated_doc
